/* 
 * Simple serial port daemon
 *
 * Usage: sportd [-f] [-D debug] -p port -d device -s speed
 *
 *	Eg. sportd -p 5000 -d /dev/ttyd0 -s 9600
 *
 * Compile: cc -o sportd sportd.c
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/select.h>

#define BSIZE    512

#define	DBG_LOGCONNECT	1
#define	DBG_LOGBYTES	2

int debug = 0;
int foreground = 0;

void
usage_exit(void)
{
	printf("Usage: sportd -p port -d device -s speed\n");
	exit(1);
}

void
connect_modem(int sockfd, char *port, int speed)
{
	int devfd;
	struct termios tty_state;
	int fdmax;
	int quit;
	fd_set readset;
	int i,n;
	char buffer[BSIZE];

	/* Open device */

	if ((devfd = open(port, O_RDWR | O_NONBLOCK)) < 0) {
		perror("open" );
		exit (1);
	}

	/* Set port parameters */

	if (tcgetattr(devfd, &tty_state) < 0) {
		perror("tcgetattr");
		exit(1);
	}
	
	tty_state.c_lflag = NOFLSH;
	tty_state.c_iflag = 0;
	tty_state.c_oflag = 0;
	tty_state.c_cflag = CS8|CREAD|CLOCAL;
	cfsetispeed (&tty_state, B9600);
	cfsetospeed (&tty_state, B9600);
	
	tty_state.c_cc[VMIN]  = 1;
	tty_state.c_cc[VTIME] = 0;
	
	if (tcsetattr(devfd, TCSAFLUSH, &tty_state) < 0)
		return ;
	
	fdmax = sockfd;
	if (devfd > fdmax) fdmax = devfd;

	/* Main relay loop */

	quit = 0;
	while (!quit)
	{
		FD_ZERO (&readset);
		FD_SET (devfd, &readset);
		FD_SET (sockfd, &readset);
		select (fdmax + 1, &readset, NULL, NULL, NULL);

		if (FD_ISSET(devfd, &readset)) {
			if ((n = read(devfd, buffer, BSIZE)) > 0) {
				if (debug & DBG_LOGBYTES) {
					printf("port->network:");
					for(i=0 ; i<n ; i++)
						printf(" %02x", buffer[i]);
					printf("\n");
				}
				if (write(sockfd, buffer, n) < 0)
					exit (1);
			}
		}

		if (FD_ISSET(sockfd, &readset)) {
			if ((n = read(sockfd, buffer, BSIZE)) > 0) {
				if (debug & DBG_LOGBYTES) {
					printf("network->port:");
					for(i=0 ; i<n ; i++)
						printf(" %02x", buffer[i]);
					printf("\n");
				}
				if (write(devfd, buffer, n) < 0)
					exit (1);
			}
			else quit = 1;
		}
	}

	close(devfd);
}

void
daemonize(void) {
	int fd;
	switch (fork()) {
	case 0:
		break;
	case -1:
		perror("fork");
		exit(1);
	default:
		_exit(0);
	}

	if (setsid() < 0) {
		perror("setsid");
		exit(1);
	}

	switch (fork()) {
	case 0:
		break;
	case -1:
		perror("fork");
		exit(1);
	default:
		_exit(0);
	}

	chdir("/");

	fd = open("/dev/null", O_RDONLY);
	if (fd != 0) {
		dup2(fd, 0);
		close(fd);
	}
	fd = open("/dev/null", O_WRONLY);
	if (fd != 1) {
		dup2(fd, 1);
		close(fd);
	}
	fd = open("/dev/null", O_WRONLY);
	if (fd != 2) {
		dup2(fd, 2);
		close(fd);
	}
}

int
main(int argc, char **argv)
{
	char *device;
	int port,speed;
	int opt;
	int servfd,sockfd;
	unsigned int on,len;
	struct sockaddr_in serv_addr,peer;

	extern char *optarg;

	/* Get parameters */
	device = NULL;
	port = speed = 0;
	while((opt=getopt(argc, argv, "d:D:fp:s:")) != -1) {
		switch(opt) {
                case 'd':
			device = optarg;
			break;
                case 'D':
			debug = atoi(optarg);
			break;
                case 'f':
			foreground =  1;
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 's':
			speed = atoi(optarg);
			break;
		default:
			usage_exit();
		}
	}

	/* Sanity check */
	if (device == NULL || port == 0 || speed == 0)
		usage_exit();

	/* Go daemon ... */
	if (foreground == 0) {
		daemonize();
		/* No debug output if in background */
		debug = 0;
	}

	/* Start listening for connections */
	if ((servfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	on = 1;
	if (setsockopt(servfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
			perror("setsockopt");
			exit(1);
	}

	bzero ((char *)&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(port);
	if (bind(servfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
		perror("bind");
		exit(1);
	}

	if (listen(servfd, 5) < 0) {
		perror("bind");
		exit(1);
	}
	
	/* Wait for connection */
	for(;;) {
		/* Get remote filedescriptor */
		len = sizeof(peer);
		if ((sockfd = accept(servfd, (struct sockaddr *) &peer, &len)) < 0) {
			perror("accept");
			exit(1);
		}
		if (debug & DBG_LOGCONNECT)
			printf("Connect from %s:%d\n",
				inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

		/* Open port and relay */
		connect_modem(sockfd,device,speed);
		if (debug & DBG_LOGCONNECT)
			printf("Connection closed\n");

		/* Done */
		close(sockfd);
	}

	/* All done */
	close(servfd);
	return 0;
}
